import warnings

from pprint import pprint
from pathlib import Path
from tqdm import tqdm

import numpy as np
import pandas as pd
from pandas.core.common import SettingWithCopyWarning
warnings.simplefilter(action="ignore", category=SettingWithCopyWarning)

import dolphindb as ddb
import dolphindb.settings as keys

import sqlalchemy as sa

from DDBLoader import DDBLoader


class DDBPITLoader(DDBLoader):

    ddb_path = "dfs://pit_stock_ts"
    ddb_dbname = "ddb_pit_stock_ts"

    num_code_partition = 50

    table_name_mapping = {
        #'CBS_AFTER_ADJ' : 'bs_common_adj',
        #'CBS_BEFORE_ADJ' : 'bs_common_ori',
        #'CCFS_AFTER_ADJ' : 'cfs_common_adj',
        #'CCFS_BEFORE_ADJ' : 'cfs_common_ori',
        #'CIS_AFTER_ADJ' : 'is_common_adj',
        #'CIS_BEFORE_ADJ' : 'is_common_ori',
        'DIV_WIND' : 'divident',
        #'EP_WIND' : 'earnings_preannouncement',
        #'PEE_WIND' : 'preliminary_earnings_estimate'
    }

    meta_col_config = {
        'WIND_CODE' : ('code', 'SYMBOL'),
        # mssql表中不需要记录的meta字段，在这里直接设置为None
        'IntCode' : None,
        'ACTUAL_ANN_DT' : None,
        'ReportPeriod' : ('report_period', 'DATE'),
        'AppearInPeriod' : ('appear_in_period', 'DATE'),
        'AppearAtDate' : ('appear_at_date', 'DATE')
    }

    date_col_set = {
        'report_period',
        'appear_in_period',
        'appear_at_date',
        'ReportPeriod',
        'AppearInPeriod',
        'AppearAtDate',
        'EQY_RECORD_DT',
        'EX_DT',
        'DVD_PAYOUT_DT',
        'S_DIV_PRELANDATE',
        'S_DIV_SMTGDATE',
        'DVD_ANN_DT',
        'S_DIV_PREANNDT'
    }

    ddb_type_mapping = {
        'float' : 'DOUBLE',
        'int' : 'INT',
        'text' : 'STRING',
        'varchar' : 'STRING',
        'str' : 'STRING'
    }

    # 基本面数据库现在存放在91服务器之上
    mssql_config = {
        'host' : '192.168.1.91',
        'username' : 'sa',
        'password' : 'xn.123',
        'dbname' : 'tr_statement'
    }


    def __init__(self):
        super().__init__()
        # 重新设定mssql_engine对象，此时我们需要使用基本面数据库
        self.mssql_engine = sa.create_engine(
            "mssql+pyodbc://{username}:{password}@{host}/{dbname}?driver=ODBC+Driver+18+for+SQL+Server".format(**self.mssql_config),
            connect_args = {
                "TrustServerCertificate": "yes"
            }, echo=False
        )


    def create_ddb_database(self):
        self.ddb_sess.run("""
            {dbName} = database(
                directory = '{dbPath}',
                partitionType = HASH,
                partitionScheme = [SYMBOL, {num_code_partition}],
                engine = 'TSDB'
            )
        """.format(
            dbName = self.ddb_dbname,
            dbPath = self.ddb_path,
            num_code_partition = self.num_code_partition
        ))


    def _make_col_config(self, mssql_table_name):
        """
        Return:
            mssql_col_name_list, ddb_col_name_list, ddb_col_type_list 
        """
        with self.mssql_engine.connect() as conn:
            col_sp_list = list(conn.execute(f"exec sp_columns {mssql_table_name}").fetchall())

        mssql_col_name_list, ddb_col_name_list, ddb_col_type_list = \
                [], [], []

        for col_sp in col_sp_list:
            _col_name = col_sp[3]
            _col_type = col_sp[5]

            # 对于meta字段，需要根据meta配置表来进行处理
            if _col_name in self.meta_col_config:
                # 跳过mssql表中 不需要记录的meta字段
                if self.meta_col_config[_col_name] is None:
                    continue
                # 字段名和字段类型都要进行映射
                mssql_col_name_list.append(_col_name)
                ddb_col_name_list.append(self.meta_col_config[_col_name][0])
                ddb_col_type_list.append(self.meta_col_config[_col_name][1])
            # 对于非meta字段，仅需要检查是否是float类型，对于float类型设置类型为DOUBLE
            else:
                # 需要之后被转换成DATE的字段，一般在原表中为为INT类型
                if _col_name in self.date_col_set:
                    mssql_col_name_list.append(_col_name)
                    ddb_col_name_list.append(_col_name)
                    ddb_col_type_list.append('DATE')
                # 按照对照表进行类型转换
                elif _col_type in self.ddb_type_mapping:
                    mssql_col_name_list.append(_col_name)
                    ddb_col_name_list.append(_col_name)
                    ddb_col_type_list.append(self.ddb_type_mapping[_col_type])
                # 对照表中没有的字段类型，就不加入到字段列表中了
                else:
                    print(f"!**Unrecognized type '{_col_type}' for column {_col_name}, will skip.")

        return mssql_col_name_list, ddb_col_name_list, ddb_col_type_list


    def create_ddb_partition_table(self, mssql_table_name):
        """创建分区表"""
        memory_table_name = mssql_table_name
        partition_table_name = self.table_name_mapping[mssql_table_name]

        mssql_col_name_list, ddb_col_name_list, ddb_col_type_list = \
                self._make_col_config(mssql_table_name)

        # 根据是否
        if 'appear_in_period' in ddb_col_name_list:
            compress_methods = """{
                'report_period' : 'delta', 
                'appear_in_period' : 'delta', 
                'appear_at_date' : 'delta'
            }"""
        else:
            compress_methods = """{
                'report_period' : 'delta',
                'appear_at_date' : 'delta'
            }"""

        # 因为已经根据`appear_in_period`分列了调整前和调整后，因此不需要对它再进行排序了
        sort_columns = "`code`report_period`appear_at_date"

        # 1. 先创建内存表，内存表中设定好列名和列类型
        # 2. 然后根据内存表创建分区表，设定分区列等信息
        self.ddb_sess.run("""
            {memory_table_name} = table(
                {capacity}:0,
                {column_name_list},
                [{column_type_list}]
            );

            if (existsTable("{ddb_path}", "{partition_table_name}")) {{
                dropTable({ddb_dbname}, "{partition_table_name}");
            }}

            {partition_table_name} = createPartitionedTable(
                dbHandle = {ddb_dbname},
                table = {memory_table_name},
                tableName = "{partition_table_name}",
                partitionColumns = 'code',
                compressMethods = {compress_methods},
                sortColumns = {sort_columns}
            );
        """.format(
            ddb_dbname = self.ddb_dbname,
            ddb_path = self.ddb_path,
            memory_table_name = memory_table_name,
            partition_table_name = partition_table_name,
            capacity = 10,
            column_name_list = '`' + '`'.join(ddb_col_name_list),
            column_type_list = ','.join(ddb_col_type_list),
            compress_methods = compress_methods.replace('\n', '').replace(' ', ''),
            sort_columns = sort_columns
        ))
        print('-' * 80)
        print(f"Did create parition table <{partition_table_name}>:")
        pprint(self.ddb_sess.run(f"schema({partition_table_name});"))
        return partition_table_name, mssql_col_name_list


    def create_ddb_partition_tables(self):
        for mssql_table_name in self.table_name_mapping:
            self.create_ddb_partition_table(mssql_table_name)

    
    def _dump_pit_to_ddb(self, mssql_table_name):
        print('Will work on table', mssql_table_name)
        # 返回的`mssql_col_name_list`可以用来对SQL-Server获取的dataframe进行列过滤
        partition_table_name, mssql_col_name_list = \
                self.create_ddb_partition_table(mssql_table_name)
        
        with self.mssql_engine.connect() as conn:
            stat = f"select distinct [WIND_CODE] from {mssql_table_name}"
            stock_id_list = list(conn.execute(stat).fetchall())

            with tqdm(stock_id_list) as pbar:
                for (stock_id,) in pbar:
                    pbar.set_description(f"Will work on {stock_id}") 
                    #pbar.set_description(f"Will fetch all data of {stock_id} from SQL Server")
                    stat = """
                        select * from {mssql_table_name}
                        where WIND_CODE='{stock_id}' and AppearAtDate>0
                    """.format(
                        mssql_table_name = mssql_table_name,
                        stock_id = stock_id
                    )
                    row_list = list(conn.execute(stat).fetchall())
                    num_rows = len(row_list)

                    # 因为对AppearAtDate做了过滤，所以有可能得到一个空的数据集
                    if num_rows == 0:
                        print(f"Will skip {stock_id} due to empty result.")
                        continue

                    #pbar.set_description(f"Will work on dumping job on {stock_id} of len {num_rows}")
                    # 这里需要对select语句获取的所有列进行一次过滤，以保证和partition table中的列一致
                    df = pd.DataFrame(row_list)[mssql_col_name_list]
                    # 需要把部分字段的int字段类型转换成DATE字段类型
                    for df_col in df.columns:
                        if df_col in self.date_col_set:
                            df[df_col] = DDBLoader.make_date(df[df_col])
                    # 因为在做数据库View的时候已经做过一轮转换了，所以这里就不需要再次转换了
                    #df['WIND_CODE'] = DDBLoader.tscode_to_windcode(df['WIND_CODE'])

                    self.ddb_sess.upload({mssql_table_name : df})
                    self.ddb_sess.run(f"{partition_table_name}.tableInsert({mssql_table_name})")


    def dump_pit_to_ddb(self):
        for mssql_table_name in self.table_name_mapping:
            self._dump_pit_to_ddb(mssql_table_name)


def main():

	# TODO: 
	# 可以使用`Fire`库，对函数调用再做一次封装，就可以避免每次运行不同参数时候需要修改内部多处的代码。

    # PIT基本面数据
    loader = DDBPITLoader()
    loader.create_ddb_database()
    #loader.create_ddb_partition_tables()
    loader.dump_pit_to_ddb()

    # 日频行情数据
    #loader = DDBDailyLoader()
    #loader.load_ddb_database()
    #loader.dump_daily_kline_to_ddb()


if __name__ == '__main__':
    main()



